#!/usr/bin/python

import errno
import os.path
import stat
import sys
from fuse import FUSE, Operations, FuseOSError

from security.cryptography import aes_gen_key, aes_decrypt, aes_encrypt, sha1
from security.profile import get_current_profile
from rpc.client import RPCClient, RPCError

class FileSystem(Operations):
    def __init__(self, hostname, port):
        prof = get_current_profile()
        self.rpc_client = RPCClient(hostname, port)

    def _decrypt_file_if_valid(self, encrypted_contents, key, real_path):
        tagged_contents = aes_decrypt(encrypted_contents, key)
        contents_pieces = tagged_contents.split("\n", 2)
        claimed_path = contents_pieces[0]
        claimed_hash_of_contents = contents_pieces[1]
        contents = contents_pieces[2]
        if claimed_path == real_path and claimed_hash_of_contents == sha1(contents):
            return contents
        else:
            return None

    def invoke_rpc(self, method, *args, **kwargs):
        try:
            return self.rpc_client.make_rpc_call(method, *args, **kwargs)
        except RPCError as e:
            raise FuseOSError(e.errno)

    def import_keys(self, path):
        prof = get_current_profile()
        pieces = path.split(os.path.sep)
        pieces[0] = '/'
        last_known_subpath = '/'
        key = prof.get_key_for_file(last_known_subpath)
        for ndx, piece in enumerate(pieces):
            subpath = os.path.join(*pieces[:ndx+1])
            new_key = prof.get_key_for_file(subpath) 
            if new_key is None:
                break
            else:
                last_known_subpath = subpath
                key = new_key
        while last_known_subpath != path:
            new_encrypted_piece = aes_encrypt(pieces[ndx], key)
            claimed_key_info_encrypted = self.invoke_rpc('import_key',
                                                         os.path.join(encrypt_path(last_known_subpath),
                                                                      new_encrypted_piece))
            if claimed_key_info_encrypted is None:
                return False
            claimed_key_info = prof.rsa_decrypt_user_key_info(claimed_key_info_encrypted)
            claimed_key = claimed_key_info['key']
            claimed_path = claimed_key_info['path']
            if claimed_path == os.path.join(last_known_subpath, pieces[ndx]):
                # the path matches up, server can't fake this, so key must be legit
                prof.set_key_for_file(os.path.join(last_known_subpath, pieces[ndx]), claimed_key)
                key = claimed_key
                last_known_subpath = os.path.join(last_known_subpath, pieces[ndx])
                ndx += 1
            else:
                return False
        return True

    def access(self, path, mode):
        prof = get_current_profile()
        key = prof.get_key_for_file(path)
        if key is None:
            success = self.import_keys(path)
            if not success:
                raise FuseOSError(errno.ENOENT)
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('access', encrypted_path, mode)

    def chmod(self, path, mode):
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('chmod', encrypted_path, mode)

    def chown(self, path, uid, gid):
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('chown', encrypted_path, uid, gid)

    def create(self, path, mode, fi=None):
        prof = get_current_profile()
        normpath = os.path.normpath(path)
        aes_key = aes_gen_key()
        prof.set_key_for_file(normpath, aes_key)
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('create', encrypted_path, sha1(os.path.normpath(path)), mode, prof.rsa_encrypt_user_key_info(path=normpath,
                                                                                                                            key=aes_key),
                               fi=fi)

    def flush(self, path, fh):
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('flush', encrypted_path, fh)

    def fsync(self, path, fdatasync, fh):
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('fsync', encrypted_path, fdatasync, fh)

    def getattr(self, path, fh=None):
        prof = get_current_profile()
        key = prof.get_key_for_file(path)
        if key is None:
            success = self.import_keys(path)
            if not success:
                raise FuseOSError(errno.ENOENT)
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('getattr', encrypted_path, fh=fh)

    def link(self, target, name):
        encrypted_target = encrypt_path(target)
        encrypted_name = encrypt_path(name)
        return self.invoke_rpc('link', encrypted_target, encrypted_name)

    def mkdir(self, path, mode):
        prof = get_current_profile()
        normpath = os.path.normpath(path)
        aes_key = aes_gen_key()
        prof.set_key_for_file(normpath, aes_key)
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('mkdir', encrypted_path, sha1(os.path.normpath(path)), mode, prof.rsa_encrypt_user_key_info(path=normpath, key=aes_key))

    def mknod(self, path, mode, dev):
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('mknod', encrypted_path, mode, dev)

    def open(self, path, flags):
        prof = get_current_profile()
        key = prof.get_key_for_file(path)
        if key is None:
            success = self.import_keys(path)
            if not success:
                raise FuseOSError(errno.ENOENT)
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('open', encrypted_path, flags)

    def read(self, path, length, offset, fh):
        # We can't arbitrarily index into encrypted data, so we have
        # to read() the entire file, decrypt it, and then index there.
        attrs = self.getattr(path, fh=fh)
        # Note that this is the full length of the data *on the
        # server* (i.e. the encrypted data)
        full_length = attrs['st_size']
        encrypted_path = encrypt_path(path)
        encrypted_contents = self.invoke_rpc('read', encrypted_path, length=full_length, offset=0, fh=fh)
        prof = get_current_profile()
        key = prof.get_key_for_file(os.path.normpath(path))
        contents = self._decrypt_file_if_valid(encrypted_contents, key, path)
        if contents is None:
            return "LIAR"
        else:
            return contents[offset:offset+length]

    def readdir(self, path, fh):
        prof = get_current_profile()
        key = prof.get_key_for_file(path)
        if key is None:
            success = self.import_keys(path)
            if not success:
                raise FuseOSError(errno.ENOENT)
        encrypted_path = encrypt_path(path)
        dirents = self.invoke_rpc('readdir', encrypted_path, fh)
        if type(dirents) is int:
            raise FuseOSError(dirents)
        # . and .. should definitely both be there for us to remove
        dirents.remove(".")
        dirents.remove("..")
        # create entire encrypted path in order to decrypt as a path
        normalized = map(lambda dirent: os.path.join(encrypted_path, dirent), dirents)
        decrypted_filenames = map(decrypt_path, normalized)
        localized = map(os.path.basename, decrypted_filenames)
        localized.extend([".", ".."])
        return localized

    def readlink(self, path):
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('readlink', encrypted_path)

    def release(self, path, fh):
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('release', encrypted_path, fh)

    def rename(self, old, new):
        attrs = self.getattr(old)
        full_length = attrs['st_size']
        mode = attrs['st_mode']
        if stat.S_ISDIR(mode):
            raise FuseOSError(errno.ENOSYS)
        old_fh = self.open(old, os.O_RDONLY)
        buf = self.read(old, full_length, 0, old_fh)
        new_fh = self.create(new, os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
        self.write(new, buf, 0, new_fh)
        prof = get_current_profile()
        new_key = prof.get_key_for_file(new)
        prof.rename(os.path.normpath(old), os.path.normpath(new), new_key)
        self.chmod(new, mode)
        self.unlink(old)

    def rmdir(self, path):
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('rmdir', encrypted_path)

    def statfs(self, path):
        prof = get_current_profile()
        key = prof.get_key_for_file(path)
        if key is None:
            success = self.import_keys(path)
            if not success:
                raise FuseOSError(errno.ENOENT)
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('statfs', encrypted_path)

    def symlink(self, target, name):
        encrypted_target = encrypt_path(target)
        encrypted_name = encrypt_path(name)
        return self.invoke_rpc('symlink', encrypted_target, encrypted_name)

    def truncate(self, path, length, fh=None):
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('truncate', encrypted_path, length, fh=fh)

    def unlink(self, path):
        encrypted_path = encrypt_path(path)
        retval = self.invoke_rpc('unlink', encrypted_path)
        prof = get_current_profile()
        prof.delete_key(path)
        return retval

    def utimens(self, path, times=None):
        encrypted_path = encrypt_path(path)
        return self.invoke_rpc('utimens', encrypted_path, times=times)

    def write(self, path, buf, offset, fh):
        encrypted_path = encrypt_path(path)
        prof = get_current_profile()
        key = prof.get_key_for_file(os.path.normpath(path))
        tagged_buf = path + "\n" + sha1(buf) + "\n" + buf
        encrypted_buf = aes_encrypt(tagged_buf, key)
        retval = self.invoke_rpc('write', encrypted_path, encrypted_buf, offset, fh)
        # write() will return either an errcode or the number of bytes
        # it wrote. Since the client doesn't know we're encrypting
        # stuff, we need to fake the amount of bytes we wrote.
        if retval < 0:
            return retval
        else:
            return len(buf)

def encrypt_path(path):
    normpath = os.path.normpath(path)
    pieces = list(normpath.split(os.path.sep))
    pieces[0] = "/"
    while "" in pieces:
        pieces.remove("")
    prof = get_current_profile()
    parent_dirname = "/"
    encrypted_pieces = ["/"]
    for ndx, piece in list(enumerate(pieces))[1:]:
        key_for_parent = prof.get_key_for_file(parent_dirname)
        encrypted_piece = aes_encrypt(piece, key_for_parent)
        encrypted_pieces.append(encrypted_piece)
        parent_dirname = os.path.join(*pieces[:ndx+1])
    return os.path.join(*encrypted_pieces)

def decrypt_path(path):
    normpath = os.path.normpath(path)
    encrypted_pieces = list(normpath.split(os.path.sep))
    encrypted_pieces[0] = "/"
    while "" in encrypted_pieces:
        encrypted_pieces.remove("")
    prof = get_current_profile()
    parent_dirname = '/'
    decrypted_pieces = ['/']
    for ndx, encrypted_piece in list(enumerate(encrypted_pieces))[1:]:
        key_for_parent = prof.get_key_for_file(parent_dirname)
        decrypted_piece = aes_decrypt(encrypted_piece, key_for_parent)
        decrypted_pieces.append(decrypted_piece)
        parent_dirname = os.path.join(*decrypted_pieces)
    return os.path.join(*decrypted_pieces)

def main(hostname, port, mountpoint):
    filesystem = FileSystem(hostname, port)
    FUSE(filesystem, mountpoint, foreground=True)

if __name__ == "__main__":
    hostname = sys.argv[1]
    port = int(sys.argv[2])
    mountpoint = sys.argv[3]
    main(hostname, port, mountpoint)
