import base64
import errno
import json
import socket
import threading

from security.profile import get_current_profile

class RPCClient(object):
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.profile = get_current_profile()

    def make_rpc_call(self, procname, *args, **kwargs):
        """
        This function makes an actual RPC call with a particular name,
        passing along its arguments. The RPC is a base64-encoded JSON
        string; consequently, the arguments passed need to be
        JSON-encodable.

        This function is threadsafe: each call to it creates a new
        socket, and each operation is tagged with a unique ID, so that
        if multiple threads are making RPC calls all at once, each
        will wait until it gets a response specifically to its call.
        """
        # Open a new socket for both reading and writing
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.host, self.port))
        rsock = sock.makefile('r')
        wsock = sock.makefile('w')

        # Make the request and send it
        request = base64.b64encode(json.dumps({'procname': procname,
                                               'args': args,
                                               'kwargs': kwargs}))
        signed_request = base64.b64encode(json.dumps({'request': request,
                                                      'username': self.profile.username,
                                                      'signature': self.profile.rsa_sign_string(request)}))
        wsock.write("%s\n" % (signed_request,))
        wsock.flush()

        # Read the response back and close the socket
        response = rsock.readline().strip()
        sock.close()
        
        data = json.loads(base64.b64decode(response))
        if data['status'] == 'error':
            raise RPCError(data['exception']['type'],
                           data['exception']['val'],
                           data['exception']['errno'])
        else:
            if type(data['retval']) == unicode:
                data['retval'] = str(data['retval'])
            return data['retval']

    def new_rpc_stub(self, procname):
        """
        This function just provides a simple way of creating a new RPC
        caller with a given name. This way you can do something like

            foo = rpc_client.new_rpc_stub('foo')
            foo(arg1, arg2, arg3)

        Instead of

            rpc_client.make_rpc_call('foo', arg1, arg2, arg3)
        """
        def wrapper(*args, **kwargs):
            return self.make_rpc_call(procname, *args, **kwargs)
        return wrapper


class RPCError(Exception):
    def __init__(self, exc_type, excval, err):
        Exception.__init__(self, "%s: %s" % (exc_type, excval))
        self.errno = err
