import base64
import json
import os

from cryptography import aes_gen_key_from_password, aes_gen_key, aes_encrypt, aes_decrypt, AES_ROOT_KEY, \
    rsa_gen_keypair, rsa_decrypt, rsa_encrypt, rsa_sign
from sql_backend import SQLBackend

CRYPT_FS_DIR = os.path.join(os.environ['HOME'], '.cryptfs')

class UserProfile(SQLBackend):
    """
    A profile object representing a *client's* profile (which is just
    a set of keys).
    """
    def __init__(self, username):
        self.username = username
        SQLBackend.__init__(self, os.path.join(CRYPT_FS_DIR, '%s.sql' % (username,)))

    def init_db(self):
        SQLBackend.init_db(self)
        self.execute_query("CREATE TABLE aes_keys (filename TEXT PRIMARY KEY, aes_key TEXT)")
        self.execute_query("INSERT INTO aes_keys (filename, aes_key) VALUES ('/', ?)", (AES_ROOT_KEY,))
        self.execute_query("CREATE TABLE rsa_keys (username TEXT, keypair TEXT, pubkey TEXT)")
        keypair, pubkey = rsa_gen_keypair()
        self.execute_query("INSERT INTO rsa_keys (username, keypair, pubkey) VALUES (?, ?, ?)", (self.username, keypair, pubkey))

    def is_decrypted(self):
        with open(self.db_path, 'rb') as fh:
            contents = fh.read()
        return contents.startswith('SQLite')

    def perform_aes_operation(self, password, operation):
        with open(self.db_path, 'rb') as fh:
            contents = fh.read()
        aes_key = aes_gen_key_from_password(password)
        encrypted = operation(contents, aes_key)
        with open(self.db_path, 'wb') as fh:
            fh.write(encrypted)        

    def encrypt(self, password):
        if self.is_decrypted():
            self.perform_aes_operation(password, aes_encrypt)

    def decrypt(self, password):
        if not self.is_decrypted():
            self.perform_aes_operation(password, aes_decrypt)
            if not self.is_decrypted():
                # the password was probably wrong
                # we're going to reencrypt and raise an exception
                self.perform_aes_operation(password, aes_encrypt)
                raise ValueError("Incorrect password! Undoing decryption.")

    def get_key_for_file(self, filename):
        rows = self.execute_query("SELECT aes_key FROM aes_keys WHERE filename=?", (filename,))
        try:
            aes_key = rows[0][0]
            return aes_key
        except IndexError:
            return None

    def set_key_for_file(self, filename, key):
        rows = self.execute_query("SELECT * FROM aes_keys WHERE filename=?", (filename,))
        if len(rows) == 0:
            # need to insert
            self.execute_query("INSERT INTO aes_keys (filename, aes_key) VALUES (?, ?)", (filename, key))
        else:
            # need to update
            self.execute_query("UPDATE aes_keys SET aes_key=? WHERE filename=?", (key, filename))

    def rename(self, old, new, new_key):
        self.execute_query("DELETE FROM aes_keys WHERE filename=?", (new,))
        self.execute_query("UPDATE aes_keys SET filename=?, aes_key=? WHERE filename=?", (new, new_key, old))

    def delete_key(self, filename):
        self.execute_query("DELETE FROM aes_keys WHERE filename=?", (filename,))

    def rsa_encrypt_user_key_info(self, path, key, pubkey=None):
        if pubkey is None:
            pubkey = self.rsa_get_own_pubkey()
        return base64.b64encode(rsa_encrypt(json.dumps(dict(key=key,
                                                            path=path)),
                                            pubkey))

    def rsa_decrypt_user_key_info(self, encrypted_user_key_info):
        keypair = self.rsa_get_own_keypair()
        return json.loads(rsa_decrypt(base64.b64decode(encrypted_user_key_info), keypair))

    def rsa_decrypt_string(self, encrypted_message):
	keypair = self.rsa_get_own_keypair()
	return rsa_decrypt(message, keypair)        

    def rsa_sign_string(self, message):
        keypair = self.rsa_get_own_keypair()
        return rsa_sign(message, keypair)

    def rsa_get_own_keypair(self):
        rows = self.execute_query("SELECT keypair FROM rsa_keys WHERE username=? LIMIT 1", (self.username,))
        return rows[0][0]

    def rsa_get_own_pubkey(self):
        return self.rsa_get_user_pubkey(self.username)

    def rsa_get_user_pubkey(self, username):
        rows = self.execute_query("SELECT pubkey FROM rsa_keys WHERE username=? LIMIT 1", (username,))
        pubkey = rows[0][0]
        return pubkey

    def rsa_set_user_pubkey(self, username, pubkey):
        rows = self.execute_query("SELECT * FROM rsa_keys WHERE username=?", (username,))
        if len(rows) == 0:
            # need to insert
            self.execute_query("INSERT INTO rsa_keys (username, pubkey) VALUES (?, ?)", (username, pubkey))
        else:
            # need to update
            self.execute_query("UPDATE rsa_keys SET pubkey=? WHERE username=?", (pubkey, username))

def get_current_username():
    filename = os.path.join(CRYPT_FS_DIR, "current")
    if os.path.exists(filename):
        with open(filename, 'r') as fh:
            current = fh.read().strip()
        if current == "":
            return None
        else:
            return current
    else:
        return None

def set_current_username(username):
    filename = os.path.join(CRYPT_FS_DIR, "current")
    if username != "":
        with open(filename, 'w') as fh:
            fh.write("%s\n" % (username,))
    else:
        if os.path.exists(filename):
            os.unlink(filename)

def get_current_profile():
    current_username = get_current_username()
    if current_username is not None:
        profile = UserProfile(current_username)
        return profile
    else:
        raise Exception("NO CURRENT PROFILE SET! RUN cryptfs-login FIRST!")

def database_exists(username):
    filename = os.path.join(CRYPT_FS_DIR, "%s.sql" % (username,))
    return os.path.exists(filename)
