#!/usr/bin/python

import base64
import errno
import json
import os
import sys

from rpc.server import RPCServer
from security.cryptography import rsa_encrypt, AES_ROOT_KEY
from security.hashpathstore import HashPathStore
from security.keystore import UserKeyStore
from security.permission import PERMISSION_READ, PERMISSION_WRITE, PERMISSION_ADMIN
from security.permissionstore import UserPermissionStore

class FileServer(object):
    def __init__(self, root, user_key_store, user_permission_store, hashpath_store):
        self.root = root
        self.user_key_store = user_key_store
        self.user_permission_store = user_permission_store
        self.hashpath_store = hashpath_store

    # Helpers
    # =======

    def _full_path(self, partial):
        """
        This method returns the full path to a given file *relative
        the FUSE mountpoint*.
        """
        if partial.startswith("/"):
            partial = partial[1:]
        path = os.path.join(self.root, partial)
        return path

    def _create_and_grant_permissions(self, path, creating_func, user_key_info_encrypted, rpc_signatory):
        retval = creating_func()
        if (retval > 0 or retval is None) and rpc_signatory is not None:
            self.user_permission_store.grant_permission_to_user(rpc_signatory, PERMISSION_ADMIN.perm, os.path.normpath(path), user_key_info_encrypted)
	return retval

    # User authentication 
    # ===================

    def register_user(self, username, rsa_pubkey, rpc_signatory=None):
        self.user_key_store.register_user(username, rsa_pubkey)
        self.user_permission_store.grant_permission_to_user(username, PERMISSION_WRITE.perm, '/', 
                                                            rsa_encrypt(base64.b64encode(json.dumps(dict(key=AES_ROOT_KEY, path='/'))),
                                                                        rsa_pubkey))

    def grant_permission_to_user(self, username, permission_to_grant, path,
                                 user_key_info_encrypted, rpc_signatory=None):
	if self.user_permission_store.get_permission_for_user(rpc_signatory, path) >= PERMISSION_ADMIN:
            self.user_permission_store.grant_permission_to_user(username, permission_to_grant, path, user_key_info_encrypted)
	else:
	    return -errno.EACCES

    def import_key(self, path, rpc_signatory=None):
        return self.user_permission_store.get_key_info_for_path(rpc_signatory, os.path.normpath(path))

    # Filesystem methods
    # ==================

    def access(self, path, mode, rpc_signatory=None):
        """
        Raises an error if file cannot be access()'d. Otherwise does
        nothing.
        """
        full_path = self._full_path(path)
        if not os.access(full_path, mode):
            return -errno.EACCES

    def chmod(self, path, mode, rpc_signatory=None):
        full_path = self._full_path(path)
        return os.chmod(full_path, mode)

    def chown(self, path, uid, gid, rpc_signatory=None):
        full_path = self._full_path(path)
        return os.chown(full_path, uid, gid)

    def getattr(self, path, fh=None, rpc_signatory=None):
        full_path = self._full_path(path)
        st = os.lstat(full_path)
        return dict((key, getattr(st, key)) for key in ('st_atime', 'st_ctime',
                     'st_gid', 'st_mode', 'st_mtime', 'st_nlink', 'st_size', 'st_uid'))

    def readdir(self, path, fh, rpc_signatory=None):
        if self.user_permission_store.get_permission_for_user(rpc_signatory, path) < PERMISSION_READ:
	    return -errno.EACCES
        full_path = self._full_path(path)
        dirents = ['.', '..']
        if os.path.isdir(full_path):
            dirents.extend(os.listdir(full_path))
        return dirents

    def readlink(self, path, rpc_signatory=None):
        pathname = os.readlink(self._full_path(path))
        if pathname.startswith("/"):
            # Path name is absolute, sanitize it.
            return os.path.relpath(pathname, self.root)
        else:
            return pathname

    def mknod(self, path, mode, dev, rpc_signatory=None):
        return os.mknod(self._full_path(path), mode, dev)

    def rmdir(self, path, rpc_signatory=None):
        if self.user_permission_store.get_permission_for_user(rpc_signatory, path) < PERMISSION_WRITE:
            return -errno.EACCES
        full_path = self._full_path(path)
        return os.rmdir(full_path)

    def mkdir(self, path, hashpath, mode, user_encrypted_key_info, rpc_signatory=None):
        if self.user_permission_store.get_permission_for_user(rpc_signatory, os.path.dirname(path)) < PERMISSION_WRITE:
            return -errno.EACCES
        if self.hashpath_store.path_exists(hashpath):
            return -errno.EEXIST
        retval = self._create_and_grant_permissions(path,
                                                    lambda: os.mkdir(self._full_path(path), mode),
                                                    user_encrypted_key_info,
                                                    rpc_signatory)
        self.hashpath_store.store_path(hashpath)
        return retval

    def statfs(self, path, rpc_signatory=None):
        full_path = self._full_path(path)
        stv = os.statvfs(full_path)
        return dict((key, getattr(stv, key)) for key in ('f_bavail', 'f_bfree',
            'f_blocks', 'f_bsize', 'f_favail', 'f_ffree', 'f_files', 'f_flag',
            'f_frsize', 'f_namemax'))

    def unlink(self, path, rpc_signatory=None):
        if self.user_permission_store.get_permission_for_user(rpc_signatory, path) < PERMISSION_WRITE:
            return -errno.EACCES
        retval = os.unlink(self._full_path(path))
        self.user_permission_store.delete_key(path, rpc_signatory)
        return retval

    def symlink(self, target, name, rpc_signatory=None):
        return os.symlink(self._full_path(target), self._full_path(name))

    def link(self, target, name, rpc_signatory=None):
        return os.link(self._full_path(target), self._full_path(name))

    def utimens(self, path, times=None, rpc_signatory=None):
        return os.utime(self._full_path(path), times)

    # File methods
    # ============

    def open(self, path, flags, rpc_signatory=None):
        full_path = self._full_path(path)
        return os.open(full_path, flags)

    def create(self, path, hashpath, mode, user_encrypted_key_info, fi=None, rpc_signatory=None):
        if self.user_permission_store.get_permission_for_user(rpc_signatory, os.path.dirname(path)) < PERMISSION_WRITE:
	    return -errno.EACCES
        if self.hashpath_store.path_exists(hashpath):
            return -errno.EEXIST
        retval = self._create_and_grant_permissions(path,
                                                  lambda: os.open(self._full_path(path), os.O_WRONLY | os.O_CREAT, mode),
                                                  user_encrypted_key_info,
                                                  rpc_signatory)
        self.hashpath_store.store_path(hashpath)
        return retval

    def read(self, path, length, offset, fh, rpc_signatory=None):
        if self.user_permission_store.get_permission_for_user(rpc_signatory, path) < PERMISSION_READ:
	    return -errno.EACCES
        os.lseek(fh, offset, os.SEEK_SET)
        return os.read(fh, length)

    def write(self, path, buf, offset, fh, rpc_signatory=None):
        if self.user_permission_store.get_permission_for_user(rpc_signatory, path) < PERMISSION_WRITE:
            return -errno.EACCES
        os.lseek(fh, offset, os.SEEK_SET)
        return os.write(fh, buf)

    def truncate(self, path, length, fh=None, rpc_signatory=None):
        if self.user_permission_store.get_permission_for_user(rpc_signatory, path) < PERMISSION_WRITE:
            return -errno.EACCES
        full_path = self._full_path(path)
        with open(full_path, 'r+') as f:
            f.truncate(length)

    def flush(self, path, fh, rpc_signatory=None):
        return os.fsync(fh)

    def release(self, path, fh, rpc_signatory=None):
        return os.close(fh)

    def fsync(self, path, fdatasync, fh, rpc_signatory=None):
        return self.flush(path, fh)

def main(hostname, port, root):
    user_key_store = UserKeyStore()
    user_permission_store = UserPermissionStore()
    hashpath_store = HashPathStore()

    fileserver = FileServer(root, user_key_store, user_permission_store, hashpath_store)

    RPCServer.register_handler('register_user', fileserver.register_user)
    RPCServer.register_handler('grant_permission_to_user', fileserver.grant_permission_to_user)
    RPCServer.register_handler('import_key', fileserver.import_key)
    RPCServer.register_handler('access', fileserver.access)
    RPCServer.register_handler('chmod', fileserver.chmod)
    RPCServer.register_handler('chown', fileserver.chown)
    RPCServer.register_handler('create', fileserver.create)
    RPCServer.register_handler('flush', fileserver.flush)
    RPCServer.register_handler('fsync', fileserver.fsync)
    RPCServer.register_handler('getattr', fileserver.getattr)
    RPCServer.register_handler('link', fileserver.link)
    RPCServer.register_handler('mkdir', fileserver.mkdir)
    RPCServer.register_handler('mknod', fileserver.mknod)
    RPCServer.register_handler('open', fileserver.open)
    RPCServer.register_handler('read', fileserver.read)
    RPCServer.register_handler('readdir', fileserver.readdir)
    RPCServer.register_handler('readlink', fileserver.readlink)
    RPCServer.register_handler('release', fileserver.release)
    RPCServer.register_handler('rmdir', fileserver.rmdir)
    RPCServer.register_handler('statfs', fileserver.statfs)
    RPCServer.register_handler('symlink', fileserver.symlink)
    RPCServer.register_handler('truncate', fileserver.truncate)
    RPCServer.register_handler('unlink', fileserver.unlink)
    RPCServer.register_handler('utimens', fileserver.utimens)
    RPCServer.register_handler('write', fileserver.write)

    RPCServer.start(hostname, port, user_key_store)

if __name__ == "__main__":
    hostname = sys.argv[1]
    port = int(sys.argv[2])
    root = sys.argv[3]
    main(hostname, port, root)
